/**
 * A cell from a maze solver.
 *
 * @author Jonathan Heathcote
 */
public class Cell
{
  // Options
  // Draw lines which are not highlighted (i.e. failed search paths)
  private static boolean drawNonHighlightedLines = false;
  
  
  // Constants defining the charachters used in input files
  public final static char ASCII_BORDER   = '#'; // Borders and walls
  public final static char ASCII_PATH     = ' '; // Explorable areas
  public final static char ASCII_ENTRANCE = '?'; // Starting points
  public final static char ASCII_EXIT     = '!'; // Target locations
  
  // Constants containing appropriate VT-100 escape sequences
  private final static String CLR_ESC                 = "\033[";
  private final static String CLR_SEP                 = ";";
  private final static String CLR_END                 = "m";
  private final static String CLR_PATH_BG             = "40";
  private final static String CLR_BORDER_BG           = "47";
  private final static String CLR_PATH_HIGHLIGHED     = "1;31";
  private final static String CLR_PATH_NORMAL         = "33";
  private final static String CLR_ENTRANCE_HIGHLIGHED = "31";
  private final static String CLR_ENTRANCE_NORMAL     = "34";
  private final static String CLR_EXIT_HIGHLIGHED     = "31";
  private final static String CLR_EXIT_NORMAL         = "31";
  private final static String CLR_RESET               = "\033[0m";
  
  // Constant representing the distance of a cell who's distance has not been
  // defined.
  public final static int UNDEFINED_DISTANCE = -1;
  
  // A refrence to the maze this cell belongs to
  private final Cell maze[][];
  
  // The coordinates of this cell in the maze
  private final int x;
  private final int y;
  
  // The primative properties of this cell
  // Is this cell a wall or border
  private boolean border;
  // Is this cell an entry point on the maze
  private boolean entrance;
  // Is this cell an exit point on the maze
  private boolean exit;
  // Is this cell highlighted (i.e. part of the shortest path)
  private boolean highlight;
  // An integer which represents the least a cell searcher has traveled to
  // get to this cell at the current point in time.
  private int distance = UNDEFINED_DISTANCE;
  
  /**
   * Constructor with type option.
   * 
   * A constructor which also accpets an ascii representation of the type of
   * cell this is.
   * 
   * @param mazeRefrence A refrence to a 2D array of cells which (will) contain
   *   this Cell.
   * @param cellX The x coordinate of this cell in the maze
   * @param cellY The y coordinate of this cell in the maze
   * @param asciiCellRepresentation A representation of this cell as a
   *   charachter.
   * 
   * @author Jonathan Heathcote
   */
  public Cell(Cell[][] mazeRefrence,
              int cellX,
              int cellY,
              char asciiCellRepresentation)
  {
    // Set instance constants
    maze = mazeRefrence;
    x = cellX;
    y = cellY;
    
    // Set up the cell based on the type of cell it meant to be
    if (asciiCellRepresentation == ASCII_BORDER)
    {
      border = true;
      entrance = false;
      exit = false;
    } // if
    else if (asciiCellRepresentation == ASCII_PATH)
    {
      border = false;
      entrance = false;
      exit = false;
    } // else if
    else if (asciiCellRepresentation == ASCII_ENTRANCE)
    {
      border = false;
      entrance = true;
      exit = false;
      
      // The cell is the entrance so therefore it must be zero distance from the
      // start.
      distance = 0;
    } // else if
    else if (asciiCellRepresentation == ASCII_EXIT)
    {
      border = false;
      entrance = false;
      exit = true;
    } // else if
  } // Cell
  
  /**
   * Is this cell a border piece.
   * @return Boolean: Is the cell a border?
   */
  public boolean isBorder()
  {
    return border;
  } // isBorder
  
  /**
    * Turn this cell into a border.
   */
  public void setToBorder()
  {
    border = true;
    entrance = false;
    exit = false;
  } // setToBorder
  
  /**
   * Is this cell an entry point?
   * @return Boolean: Is the cell an entry point?
   */
  public boolean isEntrence()
  {
    return entrance;
  } // isEntrence
  
  /**
   * Is this cell an exit point?
   * @return Boolean: Is the cell an exit point?
   */
  public boolean isExit()
  {
    return exit;
  } // isExit
  
  /**
   * Is this cell a path (ie not an entry, exit or border)?
   * @return Boolean: Is the cell a path?
   */
  public boolean isPath()
  {
    return !border && !entrance && !exit;
  } // isPath
  
  /**
   * Is this cell a search path (ie linked to others due to a search)
   * @return Boolean: Is the cell a search path?
   */
  public boolean isSearchPath()
  {
    return !border && distance != UNDEFINED_DISTANCE;
  } // isSearchPath
  
  /**
   * Find out if a path is connected to this one.
   * Does not test position, just relative search position.
   * @param otherCell Another cell
   * @return Boolean: Is the cell connected?
   */
  public boolean isConnectedPath(Cell otherCell)
  {
    // If non-highlighted paths are not being drawn return faliure if the path
    // is not highlighted
    if (!drawNonHighlightedLines && !otherCell.isHighlighted())
      return false;
    
    // If the other cell is not a searched path then it is definately not
    // connected to this one.
    if (!(!otherCell.isBorder() && otherCell.isSearchPath()))
      return false;
    // If the two cells are consecutive in the search order then they are
    // connected.
    else if (otherCell.getDistance() == getDistance() + 1
             || otherCell.getDistance() == getDistance() - 1)
      return true;
    else
      return false; // Something has gone wrong
  } // isConnectedPath
  
  /**
   * Is this cell part of the shortest path (ie highlighted)?
   * @return Boolean: is the cell to be highlighted.
   */
  public boolean isHighlighted()
  {
    return highlight;
  } // isHighlighted
  
  /**
   * Sets the state of this cell's highlighting.
   * @param highlighted Highlight this cell?
   */
  public void setHighlight(boolean highlighted)
  {
    highlight = highlighted;
  } // setHighlight
  
  /**
   * The least distance traveled by a searcher to reach this cell or
   * UNDEFINED_DISTANCE.
   * @return Distance of this cell from the nearest start point.
   */
  public int getDistance()
  {
    return distance;
  } // getDistance
  
  /**
   * Set the least distance traveled by a searcher to reach this cell or
   * UNDEFINED_DISTANCE.
   * @param newDistance Distance of this cell from the nearest start point.
   */
  public void setDistance(int newDistance)
  {
    distance = newDistance;
  } // setDistance
  
  /**
   * If a given set of coordinates is in the maze return a refrence to its
   * object, if not return null.
   * 
   * @param requestedX The x coordinate
   * @param requestedY The y coordinate
   * 
   * @return A refrence to a Cell in the maze.
   * 
   * @author Jonathan Heathcote
   */
  private Cell safeGetCellAtCoordinates(int requestedX, int requestedY)
  {
    // Note: if the maze has no rows (which should be impossible as mazes are
    // padded with a border) this will fail.
    if (requestedX >= 0 && requestedY >= 0
        && requestedY < maze[0].length && requestedX < maze.length)
      return maze[requestedX][requestedY];
    else
      return null;
  } // safeGetCellAtCoordinates
  
  /**
   * Returns a refrence to the cell on the left of this one (or null if there
   * is no cell).
   * @return Cell refrence.
   */
  public Cell leftCell()
  {
    return safeGetCellAtCoordinates(x-1, y);
  } // leftCell
  
  /**
   * Returns a refrence to the cell on the right of this one (or null if there
   * is no cell).
   * @return Cell refrence.
   */
  public Cell rightCell()
  {
    return safeGetCellAtCoordinates(x+1, y);
  } // rightCell
  
  /**
   * Returns a refrence to the cell above this one (or null if there is no
   * cell).
   * @return Cell refrence.
   */
  public Cell upCell()
  {
    return safeGetCellAtCoordinates(x, y-1);
  } // upCell
  
  /**
   * Returns a refrence to the cell below this one (or null if there is no
   * cell).
   * @return Cell refrence.
   */
  public Cell downCell()
  {
    return safeGetCellAtCoordinates(x, y+1);
  } // downCell
  
  /**
   * Convert this cell into a styalised string for printing in a maze.
   * 
   * @return A unicode and VT-100 escaped string representing this cell.
   */
  public String toString()
  {
    if (isBorder())
      // This cell is part of a border
      return CLR_ESC + CLR_BORDER_BG + CLR_END + " " + CLR_RESET;
    else if (isPath() && !isSearchPath())
      // This cell is part of the path but has not been searched
      return CLR_ESC + CLR_PATH_BG + CLR_END + " " + CLR_RESET;
    else if (isPath() && isSearchPath())
    {
      // if drawing of non-highlighted lines is disabled simply output a blank
      // path cell.
      if (!drawNonHighlightedLines && !isHighlighted())
        return CLR_ESC + CLR_PATH_BG + CLR_END + " " + CLR_RESET;
      
      // This cell is part of the path and it has been searched. Link it up with
      // other search paths around it.
      // Determine on which sides this path segment joins other search paths
      boolean joinsLeft  = isConnectedPath(leftCell());
      boolean joinsRight = isConnectedPath(rightCell());
      boolean joinsUp    = isConnectedPath(upCell());
      boolean joinsDown  = isConnectedPath(downCell());
      
      // Create an appropriate glyph for the cell depeding on what joins it
      String glyph;
      
      // Available Glyphs:
      //   ┌┐│├┤┼╴╶
      //   └┘─┬┴○╵╷
      
      // Split it to groups of 4 for countability.
      if      ( joinsLeft && joinsRight && joinsUp && joinsDown)
        glyph = "┼";
      else if ( joinsLeft && joinsRight && joinsUp &&!joinsDown)
        glyph = "┴";
      else if ( joinsLeft && joinsRight &&!joinsUp && joinsDown)
        glyph = "┬";
      else if ( joinsLeft && joinsRight &&!joinsUp &&!joinsDown)
        glyph = "─";
      
      else if ( joinsLeft &&!joinsRight && joinsUp && joinsDown)
        glyph = "┤";
      else if ( joinsLeft &&!joinsRight && joinsUp &&!joinsDown)
        glyph = "┘";
      else if ( joinsLeft &&!joinsRight &&!joinsUp && joinsDown)
        glyph = "┐";
      else if ( joinsLeft &&!joinsRight &&!joinsUp &&!joinsDown)
        glyph = "╴";
      
      else if (!joinsLeft && joinsRight && joinsUp && joinsDown)
        glyph = "├";
      else if (!joinsLeft && joinsRight && joinsUp &&!joinsDown)
        glyph = "└";
      else if (!joinsLeft && joinsRight &&!joinsUp && joinsDown)
        glyph = "┌";
      else if (!joinsLeft && joinsRight &&!joinsUp &&!joinsDown)
        glyph = "╶";
      
      else if (!joinsLeft &&!joinsRight && joinsUp && joinsDown)
        glyph = "│";
      else if (!joinsLeft &&!joinsRight && joinsUp &&!joinsDown)
        glyph = "╵";
      else if (!joinsLeft &&!joinsRight &&!joinsUp && joinsDown)
        glyph = "╷";
      else
        glyph = "?"; // Something went wrong!
      
      if (isHighlighted())
        return CLR_ESC + CLR_PATH_BG + CLR_SEP
               + CLR_PATH_HIGHLIGHED + CLR_END + glyph + CLR_RESET;
      else
        return CLR_ESC + CLR_PATH_BG + CLR_SEP
               + CLR_PATH_NORMAL + CLR_END + glyph + CLR_RESET;
    } // else if
    else if (isEntrence())
    {
      // This cell is a starting point
      // Determine on which sides this starting point has borders on
      boolean joinsLeft  = leftCell().isBorder();
      boolean joinsRight = rightCell().isBorder();
      boolean joinsUp    = upCell().isBorder();
      boolean joinsDown  = downCell().isBorder();
      
      // Create an appropriate glyph for the cell depeding on what joins it
      String glyph;
      
      // Available Glyphs:
      //   ▄ ▐ ▛ ▜ █
      //   ▀ ▌ ▙ ▟ ●
      
      // Split it to groups of 4 for countability.
      if      ( joinsLeft && joinsRight && joinsUp && joinsDown)
        glyph = "*";
      else if ( joinsLeft && joinsRight && joinsUp &&!joinsDown)
        glyph = "▼";
      else if ( joinsLeft && joinsRight &&!joinsUp && joinsDown)
        glyph = "▲";
      else if ( joinsLeft && joinsRight &&!joinsUp &&!joinsDown)
        glyph = "█";
      
      else if ( joinsLeft &&!joinsRight && joinsUp && joinsDown)
        glyph = "▶";
      else if ( joinsLeft &&!joinsRight && joinsUp &&!joinsDown)
        glyph = "█";
      else if ( joinsLeft &&!joinsRight &&!joinsUp && joinsDown)
        glyph = "█";
      else if ( joinsLeft &&!joinsRight &&!joinsUp &&!joinsDown)
        glyph = "█";
      
      else if (!joinsLeft && joinsRight && joinsUp && joinsDown)
        glyph = "◀";
      else if (!joinsLeft && joinsRight && joinsUp &&!joinsDown)
        glyph = "█";
      else if (!joinsLeft && joinsRight &&!joinsUp && joinsDown)
        glyph = "█";
      else if (!joinsLeft && joinsRight &&!joinsUp &&!joinsDown)
        glyph = "█";
      
      else if (!joinsLeft &&!joinsRight && joinsUp && joinsDown)
        glyph = "█";
      else if (!joinsLeft &&!joinsRight && joinsUp &&!joinsDown)
        glyph = "█";
      else if (!joinsLeft &&!joinsRight &&!joinsUp && joinsDown)
        glyph = "█";
      else if (!joinsLeft &&!joinsRight &&!joinsUp &&!joinsDown)
        glyph = "◆";
      else
        glyph = "?"; // Something went wrong!
      
      if (isHighlighted())
        return CLR_ESC + CLR_BORDER_BG + CLR_SEP
               + CLR_ENTRANCE_HIGHLIGHED + CLR_END + glyph + CLR_RESET;
      else
        return CLR_ESC + CLR_BORDER_BG + CLR_SEP
               + CLR_ENTRANCE_NORMAL + CLR_END + glyph + CLR_RESET;
    } // else if
    else if (isExit())
    {
      // This cell is a finishing point
      // Determine on which sides this end point has borders on
      boolean joinsLeft  = leftCell().isBorder();
      boolean joinsRight = rightCell().isBorder();
      boolean joinsUp    = upCell().isBorder();
      boolean joinsDown  = downCell().isBorder();
      
      // Create an appropriate glyph for the cell depeding on what joins it
      String glyph;
      
      // Available Glyphs:
      //   ◆▶▲
      //   █◀▼

      
      // Split it to groups of 4 for countability.
      if      ( joinsLeft && joinsRight && joinsUp && joinsDown)
        glyph = "*";
      else if ( joinsLeft && joinsRight && joinsUp &&!joinsDown)
        glyph = "▄";
      else if ( joinsLeft && joinsRight &&!joinsUp && joinsDown)
        glyph = "▀";
      else if ( joinsLeft && joinsRight &&!joinsUp &&!joinsDown)
        glyph = "█";
      
      else if ( joinsLeft &&!joinsRight && joinsUp && joinsDown)
        glyph = "▐";
      else if ( joinsLeft &&!joinsRight && joinsUp &&!joinsDown)
        glyph = "▟";
      else if ( joinsLeft &&!joinsRight &&!joinsUp && joinsDown)
        glyph = "▜";
      else if ( joinsLeft &&!joinsRight &&!joinsUp &&!joinsDown)
        glyph = "█";
      
      else if (!joinsLeft && joinsRight && joinsUp && joinsDown)
        glyph = "▌";
      else if (!joinsLeft && joinsRight && joinsUp &&!joinsDown)
        glyph = "▙";
      else if (!joinsLeft && joinsRight &&!joinsUp && joinsDown)
        glyph = "▛";
      else if (!joinsLeft && joinsRight &&!joinsUp &&!joinsDown)
        glyph = "█";
      
      else if (!joinsLeft &&!joinsRight && joinsUp && joinsDown)
        glyph = "█";
      else if (!joinsLeft &&!joinsRight && joinsUp &&!joinsDown)
        glyph = "█";
      else if (!joinsLeft &&!joinsRight &&!joinsUp && joinsDown)
        glyph = "█";
      else if (!joinsLeft &&!joinsRight &&!joinsUp &&!joinsDown)
        glyph = "●";
      else
        glyph = "?"; // Something went wrong!
      
      if (isHighlighted())
        return CLR_ESC + CLR_BORDER_BG + CLR_SEP
               + CLR_EXIT_HIGHLIGHED + CLR_END + glyph + CLR_RESET;
      else
        return CLR_ESC + CLR_BORDER_BG + CLR_SEP
               + CLR_EXIT_NORMAL + CLR_END + glyph + CLR_RESET;
    } // else if
    else
      return "?"; // Something went wrong!
  } // toString
  
  public static void setDrawNonHighlightedLines(boolean value)
  {
    drawNonHighlightedLines = value;
  } // setDrawNonHighlightedLines
} // class cell
